import sys
import os
import numpy as np
from numpy import sin, cos, tan, pi
import csv
from pprint import pprint

from tklib.tkprogvars import ProgramDir, DBDir, AtomDBDir, RietanDir
from tklib.tkfile import tkFile
from tklib.tksci.tksci import a0, torad, todeg, Round, Reduce01
from tklib.tkutils import SplitFilePath, GetList, pint, pfloat
from tklib.tkre    import DelQuote, DelSpace
import tklib.tkre as tkre

from tklib.tkcrystal.tkspacegroupobject import tkSpaceGroupObject


#DBDir             = os.path.join(ProgramDir, 'Databases')

RietanProgramPath = os.path.join(RietanDir,         "RIETAN_VENUS")
RietanPath        = os.path.join(RietanProgramPath, "RIETAN64.exe")
#SPGDBPath         = os.path.join(RietanDir,         "RIETAN_VENUS", "SPGRA")
#SPGRDBPath        = os.path.join(RietanDir,         "RIETAN_VENUS", "spgr.daf")
#ASFDCPath         = os.path.join(RietanDir,         "RIETAN_VENUS", "asfdc")

SPGDBPath         = os.path.join(DBDir, "spgra")
SPGRDBPath        = os.path.join(DBDir, "Spgr.daf")
ASFDCPath         = os.path.join(DBDir, "asfdc")
AtomsDBDir        = os.path.join(DBDir, "atoms")

#print("SPGDBPath=", SPGDBPath)

#=====================================
# Space group class
#=====================================

class tkSpaceGroup(tkSpaceGroupObject):
    def __init__(self, **args):
        super(tkSpaceGroup, self).__init__(**args)
        self.__Latt  = [0.0]*6  #np.empty(6)
        self.__SPGName = ''
        self.__LatticeSystem = None
        self.__LatticeAxis = None
        self.update(**args)

    def __del__(self):
        super(tkSpaceGroup, self).__del__()

    def __str__(self):
        return self.ClassPath()


    def SetP1(self):
        self.ClearSymmetryOperation();
        self.SetSPGName("P 1");
        self.SetiSPG(1);
        self.SetiSet(1);
#        self.AnalyzeTranslation();
        self.AddSymmetryOperation('x,y,z');
        self.SetLatticeSystem(latticesystem = 'triclinic');
        self.SetLatticeAxis(latticeaxis = 'triclinic');

    def SetSPGName(self, SPGName):
        self.__SPGName = tkre.DelQuote(SPGName)

    def SPGName(self):
        return self.__SPGName

    def SetiSPG(self, iSPG):
        self.__iSPG = pint(iSPG, 0)

    def iSPG(self):
        return self.__iSPG

    def SetiSet(self, iSet = 1):
         self.__iSet = pint(iSet, 1)

    def iSet(self):
        return self.__iSet

    def SetSpaceGroup(self, SPGName, iSPG, iSet = 1):
        self.__SPGName = tkre.DelQuote(SPGName)
        self.__iSPG    = pint(iSPG, 0)
        self.__iSet    = pint(iSet, 1)


    def iLaueG(self):
        return self.__iLaueG

    def iCenter(self):
        return self.__iCenter

    def LaueGroupByRietanIndex(self, idx = None):
        if idx is None or idx < 1 or idx > 15:
            return ''

        LaueGroup = [
            "Triclinic (1)",	    # 1 \
            "Monoclinic, a (2/m)",	# 2 \
            "Monoclinic, b (2/m)",	# 3 \
            "Monoclinic, c (2/m)",	# 4 \
            "Orthorhombic (mmm)",	# 5 \
            "Tetragonal (4/m)",	    # 6 \
            "Tetragonal (4/mmm)",	# 7 \
            "Trigonal, Rhombohedral (3)",  # 8 \
            "Trigonal, Hexagonal (3)",     # 9 \
            "Trigonal, Rhombohedral (3m)", # 10 \
            "Trigonal, Hexagonal (3m)",    # 11 \
            "Hexagonal (6/m)",             # 12 \
            "Hexagonal (6/mmm)", # 13 \
            "Cubic (m3)",		 # 14 \
            "Cubic (m3m)"		 # 15 \
            ]
        return LaueGroup[idx-1]

    def LatticeSystemFromLaueGroup(self, LaueG):
        if LaueG == "Trigonal, Rhombohedral (3)":
            return 'rhombohedral'
        if LaueG == "Trigonal, Hexagonal (3)":
            return 'trigonal'
        if LaueG == "Trigonal, Rhombohedral (3m)":
            return 'rhombohedral'
        if LaueG == "Trigonal, Hexagonal (3m)":
            return 'trigonal'
        LatticeSystem = DelSpace(LaueG)
        LatticeSystem = tkre.Sub(r'\s*\([^\s,]+\)', '', LatticeSystem)
        return LatticeSystem.lower()

    def ReadSpaceGroupFromSPGName(self, SPGName):
        SPGName1 = tkre.Sub(r'\s+', '', SPGName)
        count, iSPG, iSet = self.GetiSPGFromSPGName(SPGName)
        return self.ReadSpaceGroup(iSPG, iSet)

    def ReadNextSpaceGroup(self, db):
        line = db.ReadLine()
        if not line or tkre.Match('---', line):
            return None

        list = tkre.Search(r'(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S.*)\s*$', line)
        if list[6] is None:
            return None

        self.dprint(1, list)
        iSPG    = int(list[1])
        iSet    = int(list[2])
        iLaueG  = int(list[3])
        nCenter = int(list[4])
        nSym    = int(list[5])
        SPGName = list[6]
        LaueG = self.LaueGroupByRietanIndex(iLaueG)

        ICONDLine = db.ReadLine()
        ICOND = tkre.Search("^(.{2})(.{2})(.{2})(.{2})(.{2})(.{2})(.{2})"
                            +"(.{2})(.{2})(.{2})(.{2})(.{2})(.{2})(.{2})", ICONDLine)
        RCond = []
        for i in range(len(ICOND)-1):
             RCond.append(int(ICOND[i+1]))

        SymOp = []
        for i in range(nSym):
            op = db.ReadLine()
            SymOp.append(DelQuote(op))

        spg = {
            'SPGName': SPGName,
            'iSPG'   : iSPG,
            'iSet'   : iSet,
            'iLaueG' : iLaueG,
            'LaueG'  : LaueG,
            'LatticeSystem': self.LatticeSystemFromLaueGroup(LaueG),
#            'LatticeAxis'  : self.AnalyzeLatticeAxis(),
            'nCenter': nCenter,
            'nSym'   : nSym,
            'SymOp'  : SymOp,
            'RCond'  : RCond
            }

        return spg


    def ReadSpaceGroup(self, ispg, iset, SettoSelf = 0, dbpath = SPGDBPath, StopByError = 1):
#        print "SPGDBPath: $SpaceGroupDBPath\n" if($IsPrint);

        if ispg == 0:
            return None

        db = tkFile(dbpath, 'r')
        if not db:
            print("Error in tkcrystal.tkspacegroup.ReadSpaceGroup: "
                + "Can not read from [{}].".format(dbpath))
            if StopByError:
                exit()
            return 0

        while 1:
            spg = self.ReadNextSpaceGroup(db)
            if not spg:
                db.Close()
                print("Error in tkcrystal.tkspacegroup.ReadSpaceGroup: "
                    + "Can not find SPG for [{}.{}] from [{}].".format(ispg, iset, dbpath))
                if StopByError:
                    exit()
                return None

            if spg["iSPG"] == ispg and spg["iSet"] == iset:
                db.Close()
                if SettoSelf:
                    self.SetSPGName(spg["SPGName"])
                    self.SetiSPG(spg["iSPG"])
                    self.SetiSet(spg["iSet"])
                    self.SetiLaueGroup(spg["iLaueG"])
#                    self.AnalyzeTranslation()
#                   self.ExpandSymmetryOperation($nCentr);
                else:
                    SPG = tkSpaceGroup()
                    SPG.SetSPGName(spg["SPGName"])
                    SPG.SetiSPG(spg["iSPG"])
                    SPG.SetiSet(spg["iSet"])
                    SPG.SetiLaueGroup(spg["iLaueG"])
#                    SPG.AnalyzeTranslation()
#                   SPG.ExpandSymmetryOperation($nCentr);
                    spg["SPG"] = SPG

                return spg


    def GetSpaceGroupList(self, dbpath = SPGDBPath, StopByError = 1):
        db = tkFile(dbpath, 'r')
        if not db:
            db.Close()
            print("Error in tkcrystal.tkrietan.GetSpaceGroupList: "
                + "Can not read from [{}].".format(dbpath))
            if StopByError:
                exit()
            return 0

        spglist = []
        count = 0
        while 1:
            spg = self.ReadNextSpaceGroup(db)
            if not spg:
                break

            count += 1
            str = "%3d-%d (%s) [%d]" % (spg["iSPG"], spg["iSet"], spg["SPGName"], count)
            spglist.append(str)

        db.Close()
        return spglist


    def GetiSPGFromSPGName(self, spgname, dbpath = SPGDBPath, StopByError = 1):
        db = tkFile(dbpath, 'r')
        if not db:
            db.Close()
            print("Error in tkcrystal.tkrietan.GetiSPGFromSPGName: "
                + "Can not read from [{}].".format(dbpath))
            if StopByError:
                exit()
            return 0

        spgname1 = tkre.Sub(r'\s+', '', spgname)
        spglist = []
        count = 0
        while 1:
            spg = self.ReadNextSpaceGroup(db)
            if not spg:
                break

            count += 1
            SPGName = tkre.Sub(r'\s+', '', spg["SPGName"])
            if SPGName.upper() == spgname1.upper():
                db.Close()
                return count, spg["iSPG"], spg["iSet"]

        db.Close()
        print("Error in tkcrystal.tkrietan.GetiSPGFromSPGName: "
            + "Can not find SPG {} from [{}].".format(spgname, dbpath))
        if StopByError:
            exit()

        return -1, -1, -1


    def ClearSymmetryOperation(self):
        self.__SymmetryOperation = []
        self.__nSymmetryOperation = len(self.__SymmetryOperation)
        self.__TranslationVector = []
        self.__TranslationVector.append([0, 0, 0])
        self.__nTranslation = len(self.__TranslationVector)

    def DoSymmetryOperation(self, iop, pos, IsReduce01 = 1):
        if pos is None:
            return None

        sym = tkre.Split(r'[\s,]+', self.SymmetryOperation(iop))
#        print("sym=", sym)
        x = pos[0]
        y = pos[1]
        z = pos[2]
        if x is None or y is None or z is None: 
            return None

#        print("sym=", sym, [x,y,z])
        x1 = eval(sym[0])
        y1 = eval(sym[1])
        z1 = eval(sym[2])
        if IsReduce01:
            return [Reduce01(x1), Reduce01(y1), Reduce01(z1)]
        else:
            return [x1, y1, z1]

    def DoVelocitySymmetryOperation(self, iop, v, IsReduce01 = 0):
        if v is None:
            return None
        return self.DoSymmetryOperation(iop, v, 0)

    """
sub DoSymmetryOperation
{
	my ($this, $iop, $x, $y, $z, $IsReduce01) = @_;
	my $nSymmetryOperation  = $this->nSymmetryOperation();
	my $iSymmetryOperation  = $iop % $nSymmetryOperation;
	my $iTranslateOperation = int($iop / $nSymmetryOperation);
	my ($tx, $ty, $tz)      = $this->TranslationVector($iTranslateOperation+1);
	my $symop = $this->SymmetryOperation($iSymmetryOperation+1);
	$symop = lc $symop;
	$symop =~ s/([xyz])/\$$1/g;
	my ($sx,$sy,$sz) = ($symop =~ /^(.*?),(.*?),(.*)$/);
#print "is=$iop: T=($tx, $ty, $tz), S=($sx, $sy, $sz)\n";
	my $xs = eval($sx)+$tx;
	my $ys = eval($sy)+$ty;
	my $zs = eval($sz)+$tz;
	if($IsReduce01) {
		$xs = Utils::Reduce01(AtomSiteObject::Round01($xs));
		$ys = Utils::Reduce01(AtomSiteObject::Round01($ys));
		$zs = Utils::Reduce01(AtomSiteObject::Round01($zs));
	}
	return ($xs,$ys,$zs);
}

sub DoVelocitySymmetryOperation
{
	my ($this, $iop, $vx, $vy, $vz) = @_;
	return (0, 0, 0) if(not defined $vx);
	my $nSymmetryOperation = $this->nSymmetryOperation();
	my $iSymmetryOperation = $iop % $nSymmetryOperation;

	my $symop = $this->SymmetryOperation($iSymmetryOperation+1);
	my ($xo,$yo,$zo) = ($symop =~ /^(.*?),\s*(.*?),\s*(.*?)$/);
	my ($x1,$y1,$z1,$t1) = &SymOpToMatrix($xo);
	my ($x2,$y2,$z2,$t2) = &SymOpToMatrix($yo);
	my ($x3,$y3,$z3,$t3) = &SymOpToMatrix($zo);
	my $vxs = $x1*$vx + $y1*$vy + $z1*$vz;
	my $vys = $x2*$vx + $y2*$vy + $z2*$vz;
	my $vzs = $x3*$vx + $y3*$vy + $z3*$vz;
	return ($vxs,$vys,$vzs);
}
    """

    def nSymmetryOperation(self):
        self.__nSymmetryOperation = len(self.__SymmetryOperation)
        return self.__nSymmetryOperation

    def nTranslation(self):
        self.__nTranslation = len(self.__TranslationVector)
        return self.__nTranslation

    def TranslationVector(self, i):
        return self.__TranslationVector[i]

    def GuessBravaisLattice(self, tollatt, tolangle, EPSCoord):
        latt = self.LatticeParameters()
        SPGName = self.SPGName()
        ls = self.LatticeSystem().lower()
        if ls == '':
            LatticeSystem = self.SetLatticeSystem(None, 1, tollatt, tolangle)
        else:
            LatticeSystem = ls

        if tkre.Match(r'\s*F', SPGName, 'i'):
            BravaisLattice = 'F'
        elif tkre.Match(r'\s*I', SPGName, 'i'):
            BravaisLattice = 'I'
        elif tkre.Match(r'\s*A', SPGName, 'i'):
            BravaisLattice = 'A'
        elif tkre.Match(r'\s*B', SPGName, 'i'):
            BravaisLattice = 'B'
        elif tkre.Match(r'\s*C', SPGName, 'i'):
            BravaisLattice = 'C'

        if BravaisLattice == '':
            asl = self.ExpandedAtomSiteListByOutputMode()
        if self.IsFaceCenteredCell(asl, EPSCoord):
            BravaisLattice = 'F'
        elif self.IsBodyCenteredCell(asl, EPSCoord):
            BravaisLattice = 'I'
        elif self.IsACenteredCell(asl, EPSCoord):
            BravaisLattice = 'A'
        elif self.IsBCenteredCell(asl, EPSCoord):
            BravaisLattice = 'B'
        elif self.IsCCenteredCell(asl, EPSCoord):
            BravaisLattice = 'C'
        else:
            BravaisLattice = 'P'

        return [LatticeSystem, BravaisLattice]


    def GetBravaisLattice(self):
        SPGName = self.SPGName()
        LaueGroup = self.LatticeSystem()
        if tkre.Match(r'\s*P', SPGName, 'i'):
            BravaisLattice = "P({})".format(LaueGroup)
        elif tkre.Match(r'\s*F', SPGName, 'i'):
            BravaisLattice = "FC({})".format(LaueGroup)
        elif tkre.Match(r'\s*I', SPGName, 'i'):
            BravaisLattice = "BC({})".format(LaueGroup)
        elif tkre.Match(r'\s*A', SPGName, 'i'):
            BravaisLattice = "A({})".format(LaueGroup)
        elif tkre.Match(r'\s*B', SPGName, 'i'):
            BravaisLattice = "B({})".format(LaueGroup)
        elif tkre.Match(r'\s*C', SPGName, 'i'):
            BravaisLattice = "C({})".format(LaueGroup)

        if BravaisLattice == 'R':
             BravaisLattice = 'R(Rhombohedral)' 

        return BravaisLattice;


    def GetSymmetryOperationList(self):
        return self.__SymmetryOperation

    def SymmetryOperation(self, i):
        return self.__SymmetryOperation[i]

    def AddSymmetryOperation(self, symop):
#        symop = DelQuote(symop)
        list = tkre.Search(r'([xyz\+\-\/\d,\s]+)', symop)
        try:
            symop = list[1]
        except:
            return self.__nSymmetryOperation
#            pass
        if symop not in self.__SymmetryOperation:
            self.__SymmetryOperation.append(symop)
        self.__nSymmetryOperation = len(self.__SymmetryOperation)
        return self.__nSymmetryOperation

    def LatticeParameters(self):
        return self.__Latt

    def LatticeAxis(self):
        if self.__LatticeAxis is not None:
            return self.__LatticeAxis
        return ''

    def AnalyzeLatticeAxis(self, *, latticesystem = '', 
            SearchByLatticeParameter = 0, tollatt = 1.0e-5, tolangle = 1.0e-5):
        tolangle2 = tolangle * 1.01
        latt = self.LatticeParameters()
        if abs(latt[0] - latt[1]) < tollatt:
            if abs(latt[0] - latt[2]) < tollatt:
                if      abs(latt[3] - 90.0) < tolangle2 \
                    and abs(latt[4] - 90.0) < tolangle2 \
                    and abs(latt[5] - 90.0) < tolangle2:
                    self.__LatticeAxis = 'cubic'
                    return self.__LatticeAxis
                elif abs(latt[3] - latt[4]) < tolangle2 and abs(latt[3] - latt[5]) < tolangle2:
                    self.__LatticeAxis = 'trigonal'
                    return self.__LatticeAxis
            if      abs(latt[3] - 90.0) < tolangle2 \
                and abs(latt[4] - 90.0) < tolangle2 \
                and abs(latt[5] - 90.0) < tolangle2:
                self.__LatticeAxis = 'tetragonal'
                return self.__LatticeAxis
            elif    abs(latt[3] -  90.0) < tolangle2 \
                and abs(latt[4] -  90.0) < tolangle2 \
                and abs(latt[5] - 120.0) < tolangle2:
                self.__LatticeAxis = 'hexagonal'
                return self.__LatticeAxis

        if      abs(latt[3] - 90.0) < tolangle2 \
            and abs(latt[4] - 90.0) < tolangle2 \
            and abs(latt[5] - 90.0) < tolangle2:
            self.__LatticeAxis = 'orthorhombic'
            return self.__LatticeAxis
        elif   (abs(latt[3] - 90.0) < tolangle2 and abs(latt[4] - 90.0) < tolangle2) \
            or (abs(latt[3] - 90.0) < tolangle2 and abs(latt[5] - 90.0) < tolangle2) \
            or (abs(latt[4] - 90.0) < tolangle2 and abs(latt[5] - 90.0) < tolangle2):
            self.__LatticeAxis = 'monoclinic'
            return self.__LatticeAxis

        self.__LatticeAxis = 'triclinic'
        return self.__LatticeAxis

    def LatticeSystem(self):
        return self.__LatticeSystem
        
    def LatticeSystemFromSPGName(self, SPGName):
        if SPGName == '':
            return None
        if tkre.Search('4', SPGName) and tkre.Search('3', SPGName):
            return 'cubic'
        if tkre.Search('4', SPGName):
            return 'tetragonal'
        if tkre.Search('6', SPGName):
            return 'hexagonal'
        if tkre.Search('R', SPGName, 'i'):
            return 'rhombohedral'
        if tkre.Search('3', SPGName):
            return 'trigonal'
        if tkre.Search(r'2.*2', SPGName):
            return 'orthogonal'
        if tkre.Search(r'[mnabc].*[mnabc]', SPGName):
            return 'orthogonal'
        if tkre.Search('2', SPGName):
            return 'monoclinic'
        if tkre.Search('[mnabc]', SPGName):
            return 'monoclinic'
        return 'triclinic'

    def SetLatticeParameters(self, latt, GuessLatticeSystem = 0):
        for i in range(6):
            self.__Latt[i] = Round(latt[i], 6);
        self.SetLatticeSystem(SearchByLatticeParameter = GuessLatticeSystem)
        self.SetLatticeAxis()

    def SetLatticeSystem(self, *, latticesystem = '', 
            SearchByLatticeParameter = 0, tollatt = 1.0e-5, tolangle = 1.0e-5):
        if latticesystem != '':
            self.__LatticeSystem = latticesystem
            return self.__LatticeSystem

        SPGName = self.__SPGName
        iSPG    = self.__iSPG
        iSet    = self.__iSet
        inf = self.ReadSpaceGroup(iSPG, iSet)
        if inf:
            self.__LatticeSystem = inf["LatticeSystem"]
            return self.__LatticeSystem

        latt = self.LatticeParameters()
        if latt[0] == 0.0 or latt[1] == 0.0 or latt[2] == 0.0:
            self._LatticeSystem = None
            return self.__LatticeSystem

        if SearchByLatticeParameter == 0:
            if self.__LatticeSystem != '':
                return self.__LatticeSystem
            if self.__SPGName != '':
                return self.LatticeSystemFromSPGName(SPGName)

        self._LatticeSystem = self.AnalyzeLatticeAxis(tollatt, tolangle)
        return self.__LatticeSystem

    def SetLatticeAxis(self, *, latticeaxis = '', 
            SearchByLatticeParameter = 1, tollatt = 1.0e-5, tolangle = 1.0e-5):
        if latticeaxis != '':
            self.__LatticeAxis = latticeaxis
            return self.__LatticeAxis

        latt = self.LatticeParameters()
        if latt[0] == 0.0 or latt[0] == 0.0 or latt[0] == 0.0:
            return None

        if SearchByLatticeParameter == 0:
            if self.__LatticeAxis != '':
                return self.__LatticeSystem
            if self.__SPGName != '':
                return self.__LatticeSystemFromSPGName(SPGName)

        self.__LatticeAxis = self.AnalyzeLatticeAxis(tollatt = tollatt, tolangle = tolangle)

    def iLaueGroup(self):
        return self.__iLaueG

    def SetiLaueGroup(self, iLaueG = None):
        if iLaueG is not None:
            self.__iLaueG = iLaueG
        else:
            self.__iLaueG = self.LaueGroupByRietanIndex(iLaueG);
#        self.SetLatticeSystem(self.LatticeSystemFromLaueGroup(self.__LaueG))

    def LatticeSystem(self):
        return  self.__LatticeSystem



"""

sub ConvertToFraction
{
	my ($val, $s) = @_;
	my $Threshold = 0.01;
	$s = '' if(!defined $s);

	return "" if(abs($val) < $Threshold);
	return "+$s" if(abs($val - 1) < $Threshold);
	return "-$s" if(abs($val + 1) < $Threshold);
	for(my $i = 1 ; $i <= 24 ; $i++) {
		my $v = $val*$i;
		if(abs($v - int($v)) < $Threshold) {
			$val = "$v/$i";
			$val = "+$val" if($val !~ /^-/); #if($val > 0);
			last;
		}
	}
	return "$val$s";
}

sub BuildSymmetryOperationFromMatrix
{
	my ($x1, $y1, $z1, $t1) = @_;
#print "($x1,$y1,$z1,$t1) => ";
	$x1 = &ConvertToFraction($x1, 'x');
	$y1 = &ConvertToFraction($y1, 'y');
	$z1 = &ConvertToFraction($z1, 'z');
	$t1 = &ConvertToFraction(Utils::Reduce01($t1));
	my $s = "$x1$y1$z1$t1";
	$s =~ s/^\+//;
#print "($x1,$y1,$z1,$t1):$s\n";
	return $s;
}

sub SymOpToMatrix
{
	my ($symopstr) = (@_);

	$symopstr = lc $symopstr;
	$symopstr =~ s/'//g;
	my ($x, $y, $z, $t);

	$symopstr =~ s/x/\$xx/gi;
	$symopstr =~ s/y/\$yy/gi;
	$symopstr =~ s/z/\$zz/gi;
	my $xx = 0;
	my $yy = 0;
	my $zz = 0;
	$t = eval($symopstr);
	$xx = 1;
	$x = eval($symopstr) - $t;
	$xx = 0;
	$yy = 1;
	$y = eval($symopstr) - $t;
	$yy = 0;
	$zz = 1;
	$z = eval($symopstr) - $t;
	$t = Utils::Reduce01($t);

	return ($x, $y, $z, $t);
}

#========================================================
# 一般関数
#========================================================

sub SymmetryOperationByMatrix
{
	my ($this, $i) = @_;
	my $symop = $this->{"SymmetryOperation[$i]"};
	my ($xo,$yo,$zo) = ($symop =~ /^(.*?),\s*(.*?),\s*(.*?)$/);
	my ($x1,$y1,$z1,$t1) = &SymOpToMatrix($xo);
	my ($x2,$y2,$z2,$t2) = &SymOpToMatrix($yo);
	my ($x3,$y3,$z3,$t3) = &SymOpToMatrix($zo);
	return ($x1,$y1,$z1,$t1,
		$x2,$y2,$z2,$t2,
		$x3,$y3,$z3,$t3)
}

sub ExpandSymmetryOperation
{
	my ($this,$iCenter) = @_;

	my $SPG = $this->SPGName();
	my $nSymmetryOperation = $this->nSymmetryOperation();
#print "nSymmetryOperation:$nSymmetryOperation\n";
	my $nTranslation = $this->nTranslation();
	my @list;

	for(my $is = 0 ; $is < $nSymmetryOperation ; $is++) {
		my $symop = $this->SymmetryOperation($is+1);
		my ($x1,$y1,$z1,$t1,$x2,$y2,$z2,$t2,$x3,$y3,$z3,$t3)
				= $this->SymmetryOperationByMatrix($is+1);
		my $sx = &BuildSymmetryOperationFromMatrix($x1, $y1, $z1, $t1);
		my $sy = &BuildSymmetryOperationFromMatrix($x2, $y2, $z2, $t2);
		my $sz = &BuildSymmetryOperationFromMatrix($x3, $y3, $z3, $t3);
		my $str = "$sx,$sy,$sz";
		my $IsPassed = 1;
		for(my $i = 0 ; $i < @list ; $i++) {
			if($list[$i] eq $str) {
				$IsPassed = 0;
				last;
			}
		}
		push(@list, $str) if($IsPassed);
	}

	if($iCenter) {
		my $ns = @list;
		for(my $is = 0 ; $is < $ns ; $is++) {
			my $symop = $this->SymmetryOperation($is+1);
			my ($x1,$y1,$z1,$t1,
			$x2,$y2,$z2,$t2,
			$x3,$y3,$z3,$t3)
				= $this->SymmetryOperationByMatrix($is+1);

			my $sx = &BuildSymmetryOperationFromMatrix(
					-$x1, -$y1, -$z1, -$t1);
			my $sy = &BuildSymmetryOperationFromMatrix(
					-$x2, -$y2, -$z2, -$t2);
			my $sz = &BuildSymmetryOperationFromMatrix(
					-$x3, -$y3, -$z3, -$t3);
			my $str = "$sx,$sy,$sz";
			my $IsPassed = 1;
			for(my $i = 0 ; $i < @list ; $i++) {
				if($list[$i] eq $str) {
					$IsPassed = 0;
					last;
				}
			}
			push(@list, $str) if($IsPassed);
		}
	}

	my $ns = @list;
	for(my $is = 0 ; $is < $ns ; $is++) {
			my $symop = $list[$is]; #$this->SymmetryOperation($is+1);
			my ($xo,$yo,$zo) = ($symop =~ /^(.*?),\s*(.*?),\s*(.*?)$/);
			my ($x1,$y1,$z1,$t1) = &SymOpToMatrix($xo);
			my ($x2,$y2,$z2,$t2) = &SymOpToMatrix($yo);
			my ($x3,$y3,$z3,$t3) = &SymOpToMatrix($zo);

		for(my $it = 0 ; $it < $nTranslation ; $it++) {
			my ($tx,$ty,$tz) = $this->TranslationVector($it+1);
			my $sx = &BuildSymmetryOperationFromMatrix(
					$x1, $y1, $z1, $t1+$tx);
			my $sy = &BuildSymmetryOperationFromMatrix(
					$x2, $y2, $z2, $t2+$ty);
			my $sz = &BuildSymmetryOperationFromMatrix(
					$x3, $y3, $z3, $t3+$tz);
			my $str = "$sx,$sy,$sz";
			my $IsPassed = 1;
			for(my $i = 0 ; $i < @list ; $i++) {
				if($list[$i] eq $str) {
					$IsPassed = 0;
					last;
				}
			}
			push(@list, $str) if($IsPassed);
		}
	}
	$this->{'nSymmetryOperation'} = @list;
	for(my $i = 0 ; $i < @list ; $i++) {
		my $i1 = $i+1;
		$this->{"SymmetryOperation[$i1]"} = $list[$i];
	}

	return (@list);
}

sub AnalyzeTranslation
{
	my ($this) = @_;
	
	my $SPG = &SPGName($this);
	$this->{"TranslationVector[1][x]"} = 0;
	$this->{"TranslationVector[1][y]"} = 0;
	$this->{"TranslationVector[1][z]"} = 0;

	if($SPG =~ /^'?I/i) {
		$this->{"nTranslation"} = 2;
		$this->{"TranslationVector[2][x]"} = 0.5;
		$this->{"TranslationVector[2][y]"} = 0.5;
		$this->{"TranslationVector[2][z]"} = 0.5;
		return $this->{"nTranslation"};
	}

	if($SPG =~ /^'?A/i) {
		$this->{"nTranslation"} = 2;
		$this->{"TranslationVector[2][x]"} = 0;
		$this->{"TranslationVector[2][y]"} = 0.5;
		$this->{"TranslationVector[2][z]"} = 0.5;
		return $this->{"nTranslation"};
	}

	if($SPG =~ /^'?B/i) {
		$this->{"nTranslation"} = 2;
		$this->{"TranslationVector[2][x]"} = 0.5;
		$this->{"TranslationVector[2][y]"} = 0;
		$this->{"TranslationVector[2][z]"} = 0.5;
		return $this->{"nTranslation"};
	}

	if($SPG =~ /^'?C/i) {
		$this->{"nTranslation"} = 2;
		$this->{"TranslationVector[2][x]"} = 0.5;
		$this->{"TranslationVector[2][y]"} = 0.5;
		$this->{"TranslationVector[2][z]"} = 0;
		return $this->{"nTranslation"};
	}

	if($SPG =~ /^'?F/i) {
		$this->{"nTranslation"} = 4;
		$this->{"TranslationVector[2][x]"} = 0;
		$this->{"TranslationVector[2][y]"} = 0.5;
		$this->{"TranslationVector[2][z]"} = 0.5;
		$this->{"TranslationVector[3][x]"} = 0.5;
		$this->{"TranslationVector[3][y]"} = 0;
		$this->{"TranslationVector[3][z]"} = 0.5;
		$this->{"TranslationVector[4][x]"} = 0.5;
		$this->{"TranslationVector[4][y]"} = 0.5;
		$this->{"TranslationVector[4][z]"} = 0;
		return $this->{"nTranslation"};
	}

	$this->{"nTranslation"} = 1;
	return $this->{"nTranslation"};
}

sub GetTranslationArrays
{
	my @TranslationArrays = (
		{ 'Matrix' => [0.5, 0.5, 0.0], 'Name' => 'C' },
		{ 'Matrix' => [0.5, 0.0, 0.5], 'Name' => 'B' },
		{ 'Matrix' => [0.0, 0.5, 0.5], 'Name' => 'A' },
		{ 'Matrix' => [0.5, 0.5, 0.5], 'Name' => 'I' },
		);
	
	return @TranslationArrays;
}

sub GetTranslationalVectors
{
	my ($this) = @_;

	my $SPG        = $this->GetCSpaceGroup();
	my $SPGName    = $SPG->SPGName();

	my @VT = ([  0,   0,   0]);
	if($SPGName =~ /^\s*F/i) {
		@VT = (
			[  0,   0,   0],
			[  0, 1/2, 1/2],
			[1/2,   0, 1/2],
			[1/2, 1/2,   0],
			);
	}
	elsif($SPGName =~ /^\s*I/i) {
		@VT = (
			[  0,   0,   0],
			[1/2, 1/2, 1/2],
			);
	}
	elsif($SPGName =~ /^\s*A/i) {
		@VT = (
			[  0,   0,   0],
			[  0, 1/2, 1/2],
			);
	}
	elsif($SPGName =~ /^\s*B/i) {
		@VT = (
			[  0,   0,   0],
			[1/2,   0, 1/2],
			);
	}
	elsif($SPGName =~ /^\s*C/i) {
		@VT = (
			[  0,   0,   0],
			[1/2, 1/2,   0],
			);
	}
	return \@VT;
}

sub GetMatrixConventionalToPrimitiveCell
{
	my ($this, $ReferToConventionalCell) = @_;
	my $SPGName = ($ReferToConventionalCell and $this->{ConventionalSPGName})? 
				$this->{ConventionalSPGName} : $this->SPGName();
#print "2********[$this->{ConventionalSPGName}][$SPGName]\n";

	if($SPGName =~ /^\s*F/i) {
		return [ [ 0.5, 0.5,   0], 
			 [ 0,   0.5, 0.5],
			 [ 0.5,   0, 0.5] ]
	}
	elsif($SPGName =~ /^\s*I/i) {
		return [ [-0.5, 0.5, 0.5], 
			 [ 0.5,-0.5, 0.5],
			 [ 0.5, 0.5,-0.5] ];
	}
	elsif($SPGName =~ /^\s*A/i) {
		return [ [ 1.0, 0.0, 0.0], 
			 [ 0.0, 0.5, 0.5],
			 [ 0.0,-0.5, 0.5] ];
	}
	elsif($SPGName =~ /^\s*B/i) {
		return [ [ 0.5, 0.0, 0.5], 
			 [ 0.0, 1.0, 0.0],
			 [-0.5, 0.0, 0.5] ];
	}
	elsif($SPGName =~ /^\s*C/i) {
		return [ [ 0.5, 0.5, 0.0], 
			 [-0.5, 0.5, 0.0],
			 [ 0.0, 0.0, 1.0] ];
	}
	return [ [ 1.0, 0.0, 0.0], 
		 [ 0.0, 1.0, 0.0],
		 [ 0.0, 0.0, 1.0] ];
}

sub GetMatrixConventionalToPrimitiveReciprocalCell
{
	my ($this) = @_;
	my $SPGName = $this->SPGName();

	if($SPGName =~ /^\s*F/i) {
		return [ [-1.0, 1.0, 1.0], 
			 [ 1.0,-1.0, 1.0],
			 [ 1.0, 1.0,-1.0] ];
	}
	elsif($SPGName =~ /^\s*I/i) {
		return [ [ 1.0, 1.0,   0], 
			 [ 0,   1.0, 1.0],
			 [ 1.0,   0, 1.0] ]
	}
	elsif($SPGName =~ /^\s*A/i) {
		return [ [ 1.0, 0.0, 0.0], 
			 [ 0.0, 1.0, 1.0],
			 [ 0.0,-1.0, 1.0] ];
	}
	elsif($SPGName =~ /^\s*B/i) {
		return [ [ 1.0, 0.0, 1.0], 
			 [ 0.0, 1.0, 0.0],
			 [-1.0, 0.0, 1.0] ];
	}
	elsif($SPGName =~ /^\s*C/i) {
		return [ [ 1.0, 1.0, 0.0], 
			 [-1.0, 1.0, 0.0],
			 [ 0.0, 0.0, 1.0] ];
	}
	return [ [ 1.0, 0.0, 0.0], 
		 [ 0.0, 1.0, 0.0],
		 [ 0.0, 0.0, 1.0] ];
}

sub IsFaceCenteredCell
{
	my ($this, $pAS, $eps) = @_;

	my @m = (
		[0.0, 0.5, 0.5],
		[0.5, 0.0, 0.5],
		[0.5, 0.5, 0.0],
		);

	return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsBodyCenteredCell
{
	my ($this, $pAS, $eps) = @_;

	my @m = (
		[0.5, 0.5, 0.5]
		);

	return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsACenteredCell
{
	my ($this, $pAS, $eps) = @_;

	my @m = (
		[0.0, 0.5, 0.5]
		);

	return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsBCenteredCell
{
	my ($this, $pAS, $eps) = @_;

	my @m = (
		[0.5, 0.0, 0.5]
		);

	return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsCCenteredCell
{
	my ($this, $pAS, $eps) = @_;

	my @m = (
		[0.5, 0.5, 0.0]
		);

	return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsCompatibleLatticeSystem
{
	my ($this, $CrystalLS, $SPGLS) = @_;

	if($CrystalLS =~ /cub/i) {
		return 1;
	}
	if($CrystalLS =~ /tet/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 1;
	}
	if($CrystalLS =~ /ort/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 0 if($SPGLS =~ /tet/i);
		return 1;
	}
	if($CrystalLS =~ /hex/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 0 if($SPGLS =~ /tet/i);
		return 0 if($SPGLS =~ /ort/i);
		return 1;
	}
	if($CrystalLS =~ /trig/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 0 if($SPGLS =~ /tet/i);
		return 0 if($SPGLS =~ /ort/i);
		return 1;
	}
	if($CrystalLS =~ /rho/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 0 if($SPGLS =~ /tet/i);
		return 0 if($SPGLS =~ /ort/i);
		return 1;
	}
	if($CrystalLS =~ /mon/i) {
		return 0 if($SPGLS =~ /cub/i);
		return 0 if($SPGLS =~ /tet/i);
		return 0 if($SPGLS =~ /ort/i);
		return 0 if($SPGLS =~ /hex/i);
		return 0 if($SPGLS =~ /trig/i);
		return 0 if($SPGLS =~ /rho/i);
		return 1;
	}
	return 1 if($SPGLS =~ /trig/i or $SPGLS eq '');
	return 0;
}

sub GetLatticeConversionMatrix
{
	my ($this,  $PresetRule, $DirectionSelect, $ParametersSelect) = @_;

	my $sT  = new Math::MatrixReal(3, 3);
	my $T   = new Math::MatrixReal(3, 3);
	my $tRT = new Math::MatrixReal(3, 3);
	my $RT  = new Math::MatrixReal(3, 3);
	my $tR  = new Math::MatrixReal(3, 3);

	if($PresetRule eq 'No' or $PresetRule eq 'P1') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 1,  0,  0], 
				  [ 0,  1,  0],
				  [ 0,  0,  1] ] );
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'RhombHex') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 1, -1,  0], 
				  [ 0,  1, -1],
				  [ 1,  1,  1] ] );
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'HexRhomb') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 2/3, 1/3, 1/3],
				  [-1/3, 1/3, 1/3],
				  [-1/3,-2/3, 1/3] ]);
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'HexOrtho') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 1, 0, 0], 
				  [ 1, 2, 0],
				  [ 0, 0, 1] ] );
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'FCCPrim' or $PresetRule eq 'PrimFCC') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 0.5, 0.5,   0], 
				  [ 0,   0.5, 0.5],
				  [ 0.5,   0, 0.5] ] );
		if($PresetRule eq 'PrimFCC') {
			$sT = $sT->inverse();
		}
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'BCCPrim' or $PresetRule eq 'PrimBCC') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [-0.5, 0.5, 0.5], 
				  [ 0.5,-0.5, 0.5],
				  [ 0.5, 0.5,-0.5] ] );
		if($PresetRule eq 'PrimBCC') {
			$sT = $sT->inverse();
		}
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'ACenterPrim') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 1.0, 0.0, 0.0], 
				  [ 0.0, 0.5, 0.5],
				  [ 0.0,-0.5, 0.5] ] );
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'BCenterPrim') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 0.5, 0.0, 0.5], 
				  [ 0.0, 1.0, 0.0],
				  [-0.5, 0.0, 0.5] ] );
		$sT->transpose($sT);
	}
	elsif($PresetRule eq 'CCenterPrim') {
		$sT = Math::MatrixReal->new_from_cols( 
				[ [ 0.5, 0.5, 0.0], 
				  [-0.5, 0.5, 0.0],
				  [ 0.0, 0.0, 1.0] ] );
		$sT->transpose($sT);
	}
	else {
die "Error: Invalid conversion rule [$PresetRule].\n";

		print "  Rule is specified by Matrix:\n";
		$sT->assign(1, 1, eval(&GetArg('t11')));
		$sT->assign(1, 2, eval(&GetArg('t12')));
		$sT->assign(1, 3, eval(&GetArg('t13')));
		$sT->assign(2, 1, eval(&GetArg('t21')));
		$sT->assign(2, 2, eval(&GetArg('t22')));
		$sT->assign(2, 3, eval(&GetArg('t23')));
		$sT->assign(3, 1, eval(&GetArg('t31')));
		$sT->assign(3, 2, eval(&GetArg('t32')));
		$sT->assign(3, 3, eval(&GetArg('t33')));
	}

print "Conversion rule:\n";
	my $IsChosen = 0;
	if($ParametersSelect eq 'ai') {
		if($DirectionSelect eq 'OriginalToConverted') {
			print "Real space vectors: (a'i) = (tij)(ai)\n";
			$T = $sT;
			$IsChosen = 1;
		}
		elsif($DirectionSelect eq 'ConvertedlToOriginal') {
			print "Real space vectors: (ai) = (tij)(a'i)\n";
			$T = $sT->inverse();
			$IsChosen = 1;
		}
	}
	elsif($ParametersSelect eq 'xyz') {
		if($DirectionSelect eq 'OriginalToConverted') {
			print "Real space coordinates: (x'i) = (tij)(xi)\n";
			$T->transpose($sT);
			$T = $T->inverse();
			$IsChosen = 1;
		}
		elsif($DirectionSelect eq 'ConvertedlToOriginal') {
			print "Real space vectors: (xi) = (tij)(x'i)\n";
			$T->transpose($sT);
			$IsChosen = 1;
		}
	}
	elsif($ParametersSelect eq 'a*i') {
		if($DirectionSelect eq 'OriginalToConverted') {
			print "Reciprocal space vectors: (a'*i) = (tij)(a*i)\n";
			$T->transpose($sT);
			$T = $T->inverse();
			$IsChosen = 1;
		}
		elsif($DirectionSelect eq 'ConvertedlToOriginal') {
			print "Reciprocal space vectors: (a*i) = (tij)(a'*i)\n";
			$T->transpose($sT);
			$IsChosen = 1;
		}
	}
	elsif($ParametersSelect eq 'hkl') {
		if($DirectionSelect eq 'OriginalToConverted') {
			print "Reciprocal space coordinates: (h'i) = (tij)(hi)\n";
			$T = $sT;
			$IsChosen = 1;
		}
		elsif($DirectionSelect eq 'ConvertedlToOriginal') {
			print "Reciprocal space coordinates: (hi) = (tij)(h'i)\n";
			$T = $sT->inverse();
			$IsChosen = 1;
		}
	}
	unless($IsChosen) {
print "Selected option {Parameters: $ParametersSelect} in {$DirectionSelect} direction is not supported.\n";
		return 1;
	}

#実空間座標(x,y,z)の変換行列
	$tRT->transpose($T);
	$tRT = $tRT->inverse();

#逆空間ベクトル(a*i)の変換行列: $tRT
	$RT->transpose($T);
	$RT = $RT->inverse();

#逆空間座標(h k l)の変換行列: $T
	$tR = $T;

	return ($sT, $T, $tRT, $RT, $tR);
}

sub GetBravaisToPrimitiveCellMatrix
{
	my ($this) = @_;

	my $SPG        = $this->GetCSpaceGroup();
	my $SPGName    = $SPG->SPGName();

	my $sT = new Math::MatrixReal(3, 3);
	my $T;
	if($SPGName =~ /^\s*F/i) {
		$sT = Math::MatrixReal->new_from_cols( 
			[ [ 0.5, 0.5,   0],
			  [ 0,   0.5, 0.5],
			  [ 0.5,   0, 0.5] ] );
	}
	elsif($SPGName =~ /^\s*I/i) {
		$sT = Math::MatrixReal->new_from_cols( 
			[ [-0.5, 0.5, 0.5],
			  [ 0.5,-0.5, 0.5],
			  [ 0.5, 0.5,-0.5] ] );
	}
	elsif($SPGName =~ /^\s*A/i) {
		$sT = Math::MatrixReal->new_from_cols( 
			[ [ 1.0, 0.0, 0.0],
			  [ 0.0, 0.5, 0.5],
			  [ 0.0,-0.5, 0.5] ] );
	}
	elsif($SPGName =~ /^\s*B/i) {
		$sT = Math::MatrixReal->new_from_cols( 
			[ [ 0.5, 0.0, 0.5],
			  [ 0.0, 1.0, 0.0],
			  [-0.5, 0.0, 0.5] ] );
	}
	elsif($SPGName =~ /^\s*C/i) {
		$sT = Math::MatrixReal->new_from_cols( 
			[ [ 0.5, 0.5, 0.0],
			  [-0.5, 0.5, 0.0],
			  [ 0.0, 0.0, 1.0] ] );
	}
	$T = $sT;
	$T->transpose($T);
	
	return ($sT, $T);
}

sub GetPrimitiveCell
{
	my ($this, $IsPrint) = @_;

#実空間べクトル(ai)の変換行列
	my ($sT, $T) = $this->GetBravaisToPrimitiveCellMatrix();
if($IsPrint) {
	print "Conversion matrix for real space vector: (a'i) = (Tij)(ai)\n";
	printf "   |%10.4f %10.4f %10.4f|\n", $T->element(1,1), $T->element(1,2), $T->element(1,3);
	printf "   |%10.4f %10.4f %10.4f|\n", $T->element(2,1), $T->element(2,2), $T->element(2,3);
	printf "   |%10.4f %10.4f %10.4f|\n", $T->element(3,1), $T->element(3,2), $T->element(3,3);
}
	return $this->ConvertLattice($T, 0, 1);
}

sub ConvertByMatrix
{
	my ($this, $x, $y, $z, $T) = @_;
	
	my $x1 = $T->element(1,1)*$x + $T->element(1,2)*$y + $T->element(1,3)*$z;
	my $y1 = $T->element(2,1)*$x + $T->element(2,2)*$y + $T->element(2,3)*$z;
	my $z1 = $T->element(3,1)*$x + $T->element(3,2)*$y + $T->element(3,3)*$z;
	return ($x1, $y1, $z1);
}

sub PrimitiveLatticeVectors
{
	my ($this, $UseAtomicUnit) = @_;
	my $k = ($UseAtomicUnit)? 1.0 / 0.529177 : 1.0;

	my ($ax,$ay,$az,$bx,$by,$bz,$cx,$cy,$cz) = $this->LatticeVectors($UseAtomicUnit);
	my ($sT, $T) = $this->GetBravaisToPrimitiveCellMatrix();
	my ($ax1, $ay1, $az1) = $this->ConvertByMatrix($ax, $ay, $az, $sT);
	my ($bx1, $by1, $bz1) = $this->ConvertByMatrix($bx, $by, $bz, $sT);
	my ($cx1, $cy1, $cz1) = $this->ConvertByMatrix($cx, $cy, $cz, $sT);
#print "vector ($ax1, $ay1, $az1, $bx1, $by1, $bz1, $cx1, $cy1, $cz1)\n";
#exit;
	return ($ax1, $ay1, $az1, $bx1, $by1, $bz1, $cx1, $cy1, $cz1);
}


sub SymOpToMatrix
{
	my ($op) = @_;
	return SpaceGroupObject::SymOpToMatrix($op);
}

# From struct.pl in WIEN2k package
#
#######################################
# SPACEGROUPS                         #
#######################################

sub GetWIEN2kSPG
{
	my ($spgnum, $NotDeleteiSPG) = @_;

	my @lattype = qw/P F B CXY CYZ CXZ R H 1_P1 
			 2_P-1 3_P2 3_P2 3_P2 4_P21 
			 4_P21 4_P21 5_B2 6_Pm 6_Pm 
			 6_Pm 7_Pc 7_Pa 7_Pb 7_Pb 7_Pc 
			 7_Pa 7_Pn 7_Pn 7_Pn 8_Bm 9_Bb 
			 10_P2\/m 10_P2\/m 10_P2\/m 11_P21\/m 
			 11_P21\/m 11_P21\/m 12_B2\/m 13_P2\/c 
			 13_P2\/a 13_P2\/b 13_P2\/b 13_P2\/c 
			 13_P2\/a 13_P2\/n 13_P2\/n 13_P2\/n 
			 14_P21\/c 14_P21\/a 14_P21\/b 14_P21\/b 
			 14_P21\/c 14_P21\/a 14_P21\/n 14_P21\/n 
			 14_P21\/n 15_B2\/b 16_P222 17_P2221 
			 17_P2122 17_P2212 18_P21212 18_P22121 
			 18_P21221 19_P212121 20_C2221 20_A2122 
			 20_B2212 21_C222 21_A222 21_B222 22_F222 
			 23_I222 24_I212121 25_Pmm2 25_P2mm 25_Pm2m 
			 26_Pmc21 26_P21ma 26_Pb21m 26_Pcm21 26_P21am 
			 26_Pm21b 27_Pcc2 27_P2aa 27_Pb2b 28_Pma2 
			 28_P2mb 28_Pc2m 28_Pbm2 28_P2cm 28_Pm2a 
			 29_Pca21 29_P21ab 29_Pc21b 29_Pbc21 29_P21ca 
			 29_Pb21a 30_Pnc2 30_P2na 30_Pb2n 30_Pcn2 
			 30_P2an 30_Pn2b 31_Pmn21 31_P21mn 31_Pn21m 
			 31_Pnm21 31_P21nm 31_Pm21n 32_Pba2 32_P2cb 
			 32_Pc2a 33_Pna21 33_P21nb 33_Pc21n 33_Pbn21 
			 33_P21cn 33_Pn21a 34_Pnn2 34_P2nn 34_Pn2n 
			 35_Cmm2 35_A2mm 35_Bm2m 36_Cmc21 36_A21ma 
			 36_Bb21m 36_Ccm21 36_A21am 36_Bm21b 37_Ccc2 
			 37_A2aa 37_Bb2b 38_Amm2 38_B2mm 38_Cm2m 39_Abm2 
			 39_B2cm 39_Cm2a 39_Bma2 39_C2mb 39_Ac2m 40_Ama2 
			 40_B2mb 40_Cc2m 40_Bbm2 40_C2cm 40_Am2a 41_Aba2 
			 41_B2cb 41_Cc2a 41_Bba2 41_C2cb 41_Ac2a 42_Fmm2 
			 42_F2mm 42_Fm2m 43_Fdd2 43_F2dd 43_Fd2d 44_Imm2 
			 44_I2mm 44_Im2m 45_Iba2 45_I2cb 45_Ic2a 46_Ima2 
			 46_I2mb 46_Ic2m 46_Ibm2 46_I2cm 46_Im2a 47_Pmmm 
			 48_Pnnn 49_Pccm 49_Pmaa 49_Pbmb 50_Pban 50_Pncb 
			 50_Pcna 51_Pmma 51_Pbmm 51_Pmcm 51_Pmam 51_Pmmb 
			 51_Pcmm 52_Pnna 52_Pbnn 52_Pncn 52_Pnan 52_Pnnb 
			 52_Pcnn 53_Pmna 53_Pbmn 53_Pncm 53_Pman 53_Pnmb 
			 53_Pcnm 54_Pcca 54_Pbaa 54_Pbcb 54_Pbab 54_Pccb 
			 54_Pcaa 55_Pbam 55_Pmcb 55_Pcma 56_Pccn 56_Pnaa 
			 56_Pbnb 57_Pbcm 57_Pmca 57_Pbma 57_Pcmb 57_Pcam 
			 57_Pmab 58_Pnnm 58_Pmnn 58_Pnmn 59_Pmmn 59_Pnmm 
			 59_Pmnm 60_Pbcn 60_Pnca 60_Pbna 60_Pcnb 60_Pcan 
			 60_Pnab 61_Pbca 61_Pcab 62_Pnma 62_Pbnm 62_Pmcn 
			 62_Pnam 62_Pmnb 62_Pcmn 63_Cmcm 63_Amma 63_Bbmm 
			 63_Bmmb 63_Ccmm 63_Amam 64_Cmca 64_Abma 64_Bbcm 
			 64_Bmab 64_Ccmb 64_Acam 65_Cmmm 65_Ammm 65_Bmmm 
			 66_Cccm 66_Amaa 66_Bbmb 67_Cmma 67_Abmm 67_Bmcm 
			 67_Bmam 67_Cmmb 67_Acmm 68_Ccca 68_Abaa 68_Bbcb 
			 68_Bbab 68_Cccb 68_Acaa 69_Fmmm 70_Fddd 71_Immm 
			 72_Ibam 72_Imcb 72_Icma 73_Ibca 73_Icab 74_Imma 
			 74_Ibmm 74_Imcm 74_Imam 74_Immb 74_Icmm 75_P4 
			 76_P41 77_P42 78_P43 79_I4 80_I41 81_P-4 82_I-4 
			 83_P4\/m 84_P42\/m 85_P4\/n 86_P42\/n 87_I4\/m 
			 88_I41\/a 89_P422 90_P4212 91_P4122 92_P41212 
			 93_P4222 94_P42212 95_P4322 96_P43212 97_I422 
			 98_I4122 99_P4mm 100_P4bm 101_P42cm 102_P42nm 
			 103_P4cc 104_P4nc 105_P42mc 106_P42bc 107_I4mm 
			 108_I4cm 109_I41md 110_I41cd 111_P-42m 112_P-42c 
			 113_P-421m 114_P-421c 115_P-4m2 116_P-4c2 117_P-4b2 
			 118_P-4n2 119_I-4m2 120_I-4c2 121_I-42m 122_I-42d 
			 123_P4\/mmm 124_P4\/mcc 125_P4\/nbm 126_P4\/nnc 
			 127_P4\/mbm 128_P4\/mnc 129_P4\/nmm 130_P4\/ncc 
			 131_P42\/mmc 132_P42\/mcm 133_P42\/nbc 134_P42\/nnm 
			 135_P42\/mbc 136_P42\/mnm 137_P42\/nmc 138_P42\/ncm 
			 139_I4\/mmm 140_I4\/mcm 141_I41\/amd 142_I41\/acd 
			 143_P3 144_P31 145_P32 146_R3 147_P-3 148_R-3 
			 149_P312 150_P321 151_P3112 152_P3121 153_P3212 
			 154_P3221 155_R32 156_P3m1 157_P31m 158_P3c1 
			 159_P31c 160_R3m 161_R3c 162_P-31m 163_P-31c 
			 164_P-3m1 165_P-3c1 166_R-3m 167_R-3c 168_P6 
			 169_P61 170_P65 171_P62 172_P64 173_P63 174_P-6 
			 175_P6\/m 176_P63\/m 177_P622 178_P6122 179_P6522 
			 180_P6222 181_P6422 182_P6322 183_P6mm 184_P6cc 
			 185_P63cm 186_P63mc 187_P-6m2 188_P-6c2 189_P-62m 
			 190_P-62c 191_P6\/mmm 192_P6\/mcc 193_P63\/mcm 
			 194_P63\/mmc 195_P23 196_F23 197_I23 198_P213 
			 199_I213 200_Pm-3 201_Pn-3 202_Fm-3 203_Fd-3 
			 204_Im-3 205_Pa-3 206_Ia-3 207_P432 208_P4232 
			 209_F432 210_F4132 211_I432 212_P4332 213_P4132 
			 214_I4132 215_P-43m 216_F-43m 217_I-43m 218_P-43n 
			 219_F-43c 220_I-43d 221_Pm-3m 222_Pn-3n 223_Pm-3n 
			 224_Pn-3m 225_Fm-3m 226_Fm-3c 227_Fd-3m 228_Fd-3c 
			 229_Im-3m 230_Ia-3d/; 

	for(my $i = 0 ; $i < @lattype ; $i++) {
		my $s = $lattype[$i];
		next unless($s =~ /^(\d+)_(.*)$/);
		my $num = $1;
		my $spgname = $2;
#print "$num: $spgname ($spgnum)$LF";
		next if($num != $spgnum);
		return $s if($NotDeleteiSPG);
		return $spgname;
	}
	return "NotFound";
}


sub ICONDTypeStrByRietanIndex
{
	my ($idx) = @_;

	my @ICONDTypeStr = (
		"00L", "0K0", "0KL", "H00", "H0L",
		"HK0", "HKL", "0KK", "HH0", "HHL",
		"H0H", "HKK", "HKH", "HHH"
	);

	return $ICONDTypeStr[$idx];
}

sub ICONDCondStrByRietanIndex
{
	my ($idx) = @_;

	my @ICONDCondStr = (
		"None",
		"H=2N",          "K=2N",           "L=2N",     "K+L=2N",    "H+L=2N",
		"H+K=2N",        "H,K,L ODD/EVEN", "K+L=4N",   "H+L=4N",    "H+K=4N",
		"2H+L=2N",       "2H+L=4N",        "H+K+L=2N", "-H+K+L=3N", "H-K+L=3N",
		"H=4N",          "K=4N",           "L=3N",     "L=4N",      "L=6N",
		"|H|>=|K|>=|L|", "2H+K=2N",        "2H+K=4N",  "H+2K=2N",   "H+2K=4N",
		"H=2N,K=2N",     "K=2N,L=2N",      "H=2N,L=2N"
	);

	return $ICONDCondStr[$idx];
}


sub SearchiSPGFromSPGName
{
	my ($this, $spgname, $iset) = @_;
	$spgname =~ s/\s//g;

	unless(open(IN, "<$SpaceGroupDBPath")) {
		return 0;
	}
	my $NSPGR;
	my $NSET;
	my $LAUEG;
	my $NCENTR;
	my $NSYM;
	my $SPGR;
	my $BravaisCell;
	my $nTranslate;
	my $TotalSymOperation;
	my $ICONDLine;
	my @ICOND;
	my @strICOND;
	my @SymmetryOperation;
	while(<IN>) {
		my $line = $_;
		$line =~ /^\s*?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\S.*)$/;
		$NSPGR  = $1;
		$NSET   = $2;
		$LAUEG  = $3;
		$NCENTR = $4;
		$NSYM   = $5;
		$SPGR   = $6;

		$BravaisCell = substr($SPGR, 0, 1);
		$nTranslate = 1;
		$nTranslate = 2 if($BravaisCell eq 'I');
		$nTranslate = 2 if($BravaisCell eq 'A');
		$nTranslate = 2 if($BravaisCell eq 'B');
		$nTranslate = 2 if($BravaisCell eq 'C');
		$nTranslate = 4 if($BravaisCell eq 'F');

		$TotalSymOperation = $NSYM * (1+$NCENTR) * $nTranslate;
		$ICONDLine = <IN>;
		@ICOND = unpack("a2a2a2a2a2a2a2a2a2a2a2a2a2a2", $ICONDLine);
		for(my $i = 0 ; $i < 14 ; $i++) {
			Utils::DelSpace($ICOND[$i]);
			next if($ICOND[$i] eq '' or $ICOND[$i] <= 0);
			my $s = sprintf "%3s: %s",
				&ICONDTypeStrByRietanIndex($i),
				&ICONDCondStrByRietanIndex($ICOND[$i]);
			push(@strICOND, $s);
		}
		for(my $i = 0 ; $i < $NSYM ; $i++) {
			my $SymOp = <IN>;
			Utils::DelSpace($SymOp);
			push(@SymmetryOperation, $SymOp);
		}

		my $SPGR2 = $SPGR;
		$SPGR2 =~ s/\s//g;
		if($iset and uc $spgname eq uc $SPGR2 and $NSET == $iset) {
			close(IN);
#print "S: $SPGR $NSET\n";
			return ($SPGR, $NSPGR, $NSET);
		}
		elsif(uc $spgname eq uc $SPGR) {
			close(IN);
			return ($SPGR, $NSPGR, $NSET);
		}
	}
	close(IN);

	return 1;
}

sub ReadRietanSpaceGroupDB
{
	my ($this, $spgnum, $iset) = @_;

	unless(open(IN, "<$SpaceGroupDBPath")) {
		return 0;
	}
	my $NSPGR;
	my $NSET;
	my $LAUEG;
	my $NCENTR;
	my $NSYM;
	my $SPGR;
	my $BravaisCell;
	my $nTranslate;
	my $TotalSymOperation;
	my $ICONDLine;
	my @ICOND;
	my @strICOND;
	my @SymmetryOperation;
	while(<IN>) {
		my $line = $_;
		$line =~ /^\s*?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\d+?)\s+?(\S.*)$/;
#print "spgnum: $spgnum\n";
#print "line: $line";
		$NSPGR  = $1;
		$NSET   = $2;
		$LAUEG  = $3;
		$NCENTR = $4;
		$NSYM   = $5;
		$SPGR   = $6;
#print "SPGInf: $NSPGR ($spgnum): $SPGR [$NSET ($iset)]  (nSym=$NSYM)\n";

		$BravaisCell = substr($SPGR, 0, 1);
		$nTranslate = 1;
		$nTranslate = 2 if($BravaisCell eq 'I');
		$nTranslate = 2 if($BravaisCell eq 'A');
		$nTranslate = 2 if($BravaisCell eq 'B');
		$nTranslate = 2 if($BravaisCell eq 'C');
		$nTranslate = 4 if($BravaisCell eq 'F');

		$TotalSymOperation = $NSYM * (1+$NCENTR) * $nTranslate;
		$ICONDLine = <IN>;
		@ICOND = unpack("a2a2a2a2a2a2a2a2a2a2a2a2a2a2", $ICONDLine);
		for(my $i = 0 ; $i < 14 ; $i++) {
			Utils::DelSpace($ICOND[$i]);
			next if($ICOND[$i] eq '' or $ICOND[$i] <= 0);
			my $s = sprintf "%3s: %s",
				&ICONDTypeStrByRietanIndex($i),
				&ICONDCondStrByRietanIndex($ICOND[$i]);
			push(@strICOND, $s);
		}
		@SymmetryOperation = ();
		for(my $i = 0 ; $i < $NSYM ; $i++) {
			my $SymOp = <IN>;
			Utils::DelSpace($SymOp);
			push(@SymmetryOperation, $SymOp);
#print "Check: SymOp[$i]: $SymOp\n";
		}
		if($iset > 0) {
			last if($NSPGR == $spgnum and $NSET == $iset);
		}
		else {
			last if($NSPGR == $spgnum);
		}
	}
	close(IN);

	return undef unless($NSPGR == $spgnum);

	$this->SetSPGName($SPGR);
	my $LatticeSystem = LaueGroupByRietanIndex($LAUEG);
	($LatticeSystem) 
		= Utils::Split("[,\\s]+", LaueGroupByRietanIndex($LAUEG));
	$LatticeSystem = lc $LatticeSystem;
print "LatticeSystem: $LatticeSystem\n";
	$this->SetLatticeSystem($LatticeSystem);
	$this->AnalyzeTranslation();
	for(my $i = 0 ; $i < $NSYM ; $i++) {
		$this->AddSymmetryOperation($SymmetryOperation[$i]);
print "SymOp[$i]: $SymmetryOperation[$i]\n";
	}
#print "nSym0=", $this->{'nSymmetryOperation'}, "\n";

	return 1;
}


sub SetRietanDiffractinCondition
{
	my ($this, @ICOND) = @_;
	for(my $i = 0 ; $i < @ICOND ; $i++) {
		$this->{"RietanDiffractionCondition[$i]"} = $ICOND[$i];
	}
	return;
}

sub RietanDiffractinCondition
{
	my ($this, $i) = @_;
	return $this->{"RietanDiffractionCondition[$i]"};
}

"""
