from math import exp, sqrt, log
import numpy as np
from scipy import optimize          # newton関数はscipy.optimizeモジュールに入っている
from scipy.interpolate import interp1d


from tklib.tkobject import tkObject
from tklib.tkutils import terminate
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tktransport.tkTransport import rieman, integrate_Simpson, integrate_Simpson_list, fe, fh


class tkDOS(tkObject):
    def __init__(self):
        super(tkObject, self).__init__()

        self.vasp     = None
        self.vaspinf  = None
        self.infile   = "DOSCAR"
        self.Elist    = None
        self.DOSlist  = None
        self.nE       = None
        self.Emin     = None
        self.Emax     = None
        self.Estep    = None
# Interpolation type for DOS
        self.func_dos    = None  # Interpolation function givenb y self.DOS = interp1d(self.Elist, self.DOSlist, kind = self.interp_type)
        self.interp_type = 'linear'
#       self.interp_type = 'cubic'

# Flag to specify the data to determine band edges: 'EIGENVAL', 'DOSCAR'
        self.data_for_bandedges = 'EIGENVAL'  
        self.EV  = None
        self.EC  = None
        self.NA  = None  # cm^-3
        self.ND  = None  # cm^-3
        self.dEA = None  # Acceptor level measured from EV, eV
        self.dED = None  # Donor level measured from EC, eV
        self.EA  = None  # Acceptor states are not considered if EA = None
        self.ED  = None  # Donor states are not considered if ED = None

# other parameters
        self.EF0  = 0.0    # EF at charge neutral
        self.Egth = 1.0e18 # Critera DOS value to search band edges, cm^-3

        self.T0     = None
        self.EFatT0 = 0.0    # EF at T0 [K]
        self.NeatT0 = 0.0    # Total electron number at T0 [K]
        self.dQatT0 = 0.0    # dQ = Ne + NAm - Nh - NDp at T0 [K]

# integration parameters
# Integration range w.r.t. kBT
        self.Einteg0 = -3.0
        self.Einteg1 =  3.0
        self.nrange  =  6.0
# Value when E in func_dos does not exist. Terminate if outrangeE = None 
        self.outrangeE = None

# Relative accuracy of the quad functin of scipy
        self.eps_quad = 1.0e-5
        self.nmaxdiv_quad = 150

# Bisection method parameters
# Initial search range, EPS, max iteration number, output cycle
        self.dE_bisec            = 0.4
        self.eps_bisec           = 1.0e-7
        self.nmaxiter_bisec      = 200
        self.iprintiterval_bisec = 1

    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()


    def set(self, **args):
        self.update(args)


    def set_dos(self, E_list, DOS_list, EVBM = None, ECBM =None):
        self.Elist   = E_list
        self.nE      = len(self.Elist)
        self.Emin    = min(self.Elist)
        self.Emax    = max(self.Elist)
        self.DOSlist = DOS_list
        self.func_dos = interp1d(self.Elist, self.DOSlist, kind = self.interp_type)
        if EVBM is not None:
            self.EV  = EVBM
        if ECBM is not None:
            self.EC  = ECBM

    def get_dos(self):
        return self.Elist, self.DOSlist, self.func_dos
        
    def read_vaspfiles(self, infile, EF0 = None, normalize_E = True, unit = '/cm3', data_for_bandedges = None, 
            print_level = 3, usage = None, exit_by_error = True):
        if EF0 is None:
            normalize_EF0 = False
#        else:
#            normalize_EF0 = True
        if data_for_bandedges is None:
            data_for_bandedges = self.data_for_bandedges
        else:
            self.data_for_bandedges = data_for_bandedges

        vasp = tkVASP()

        if print_level:
            print("")
            print("============ Read VASP files ===========")
        
        inf = vasp.read_files(infile, "INCAR|POSCAR|KPOINTS|OUTCAR|EIGENVAL|DOSCAR", normalize_E = normalize_E,
                    exit_by_error = exit_by_error, print_level = print_level, terminate = lambda message: terminate(message, usage),
                    )
        incarinf    = inf["INCAR"]
        cry         = inf["POSCAR"]
        outcarinf   = inf["OUTCAR"]
        eigenvalinf = inf["EIGENVAL"]
        doscarinf   = inf["DOSCAR"]

        a, b, c, alpha, beta, gamm = cry.LatticeParameters()
        self.Vcell        = cry.Volume()
#        types        = cry.AtomTypeList()
#        ntypes       = len(types)
#        sites        = cry.ExpandedAtomSiteList()
#        nsites       = len(sites)
#        atomnamelist = cry.get_atom_name_list(mode = 'all', NameOnly = True)
#        poslist      = cry.get_position_list(mode = 'all', IsReduce01 = True)
        if print_level:
            cry.PrintInf("cell")
#        print("V(cell): ", Vcell, " A^3")
#    print("Atom names: ", atomnamelist)
#    print("positions:")
#    print(poslist)

        self.EF = outcarinf["EF"]
        ISPIN   = outcarinf["ISPIN"]

        self.Elist   = doscarinf["E"]
        self.nE      = doscarinf["nE"]
        self.Emin    = min(self.Elist)
        self.Emax    = max(self.Elist)
        self.DOSlist = doscarinf["TotalDOS"]
#       tDOSup = doscarinf["TotalDOSup"]
#       Neup   = doscarinf["Neup"]
#       tDOSdn = doscarinf["TotalDOSdn"]
#       Nedn   = doscarinf["Nedn"]

#        if unit == '/cm3':
#            for i in range(self.nE):
#                self.DOSlist[i] /= self.Vcell * 1.0e-24

        if print_level:
            if normalize_E:
                print("DOS energy: EF = {} eV normalized to 0".format(self.EF))
            else:
                print("DOS energy: EF = {} eV normalized to 0".format(self.EF))
            if unit == '/cm3':
                print("DOS unit converted to cm^-3 using Vcell = {:10.6g} cm^-3".format(self.Vcell * 1.0e-24))

        if EF0 is None:
            EF0 = self.EF

        if normalize_E:
            EF0 -= self.EF
            self.EF = 0.0

        if print_level:
            print("Information in OUTCAR:")
            print("  ISPIN: ", ISPIN)
            print("  EF  = {} eV".format(self.EF))
            print("  EF0 to search band edges: {} eV".format(self.EF0))

        nk      = eigenvalinf["nk"]
        nLevels = eigenvalinf["nLevels"]
        if normalize_E:
            bandedgeinf = vasp.find_band_edges_from_eigenval(EF0 = 0.0, normalize_E = normalize_E, eigenvalinf = eigenvalinf, ISPIN = ISPIN)
        else:
            bandedgeinf = vasp.find_band_edges_from_eigenval(EF0 = EF0, normalize_E = normalize_E, eigenvalinf = eigenvalinf, ISPIN = ISPIN)
        if print_level:
            print("k points in EIGENVAL:")
            print("  nk=", nk)
            print("  nLevels=", nLevels)
#    for i in range(nk):
#        el = eigenvalinf['EList'][i]
#        kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
#        print("  ({:8.4f} {:8.4f} {:8.4f}) w={:8.4g}  {:8.4f} {:12.4f}".format(kx, ky, kz, wk, dk, ktot))
        self.EV     = bandedgeinf["EV"]
        self.EC     = bandedgeinf["EC"]
        self.Eg     = bandedgeinf["Eg"]
        self.EHOMO  = bandedgeinf["EHOMO"]
        self.ELUMO  = bandedgeinf["ELUMO"]
        self.Emin   = self.Elist[0]
        self.Emax   = self.Elist[self.nE-1]
        self.Estep  = self.Elist[1] - self.Elist[0]
#        print("EF=", outcarinf["EF"], self.EF)
#        print("EC=", bandedgeinf["EC"], self.EC)
#        print("EV=", bandedgeinf["EV"], self.EV)
#        exit()

        if print_level:
            print("")
            print("Band edges from EIGENVAL:")
            if normalize_E:
                print("    EF    = {:10.6f} => 0.0 eV".format(self.EF))
            else:
                print("    EF    = {:10.6f}".format(self.EF))
            print("    EV    = {:10.6f} eV".format(self.EV))
            print("    EC    = {:10.6f} eV".format(self.EC))
            print("    Eg    = {:10.6f} eV".format(self.Eg))

            print("")
            print("============ Calculation configuration ===========")
            print("DOS configuration")
            print("nE=", self.nE)
            print("  E range: {} - {}, {} eV step".format(self.Emin, self.Emax, self.Estep))

            print("")
            print("data_for_bandedges:", self.data_for_bandedges)

        if self.data_for_bandedges == 'DOSCAR':
            self.EV, self.EC = vasp.find_band_edges_from_dos(self.Elist, self.DOSlist, EF0 = EF0, DOSth = self.Egth)
#            self.EV, self.EC = vasp.find_band_edges_from_dos(self.Elist, self.DOSlist, EF0 = 0.0, DOSth = self.Egth)
            self.Eg = self.EC - self.EV
            if print_level:
                print("Band edges from the normalized DOSCAR:")
                print("    EF    = {:10.6f} eV".format(self.EF))
                print("    EV    = {:10.6f} eV".format(self.EV))
                print("    EC    = {:10.6f} eV".format(self.EC))
                print("    Eg    = {:10.6f} eV".format(self.Eg))

        self.ED = self.EC + self.dED
        self.EA = self.EV + self.dEA
        if print_level:
            print("  Donor level   : {:10.6g} eV, {:12.6g} cm^-3".format(self.ED, self.ND))
            print("  Acceptor level: {:10.6g} eV, {:12.6g} cm^-3".format(self.EA, self.NA))

        self.func_dos = interp1d(self.Elist, self.DOSlist, kind = self.interp_type)

        self.infile  = infile
        self.vasp    = vasp
        self.vaspinf = inf

# define the DOS function
    def DOS(self, E):
        if E < self.Emin or self.Emax < E:
            if self.outrangeE is None:
                print("Error in f_dos: Given E={} exceeds the DOS E range [{}, {}]".format(E, self.Emin, self.Emax))
                exit()
            else:
                return self.outrangeE

        return self.func_dos(E)

# Define the function to be integrated
    def DOSfe(self, E, T, EF):
        return self.DOS(E) * fe(E, T, EF)

    def DOSfh(self, E, T, EF):
        return self.DOS(E) * fh(E, T, EF)

# Calculate the number of electrons from E0 to E1 with the Fermi level EF at T
    def Ne(self, T, EF, E0, E1, Estep):
#        print("Ne: E range={:8.4g} - {:8.4g}  DOS={:12.4g}-{:12.4g}".format(E0, E1, self.DOSfe(E0, T, EF), self.DOSfe(E1, T, EF)))
        N = integrate_Simpson(lambda E: self.DOSfe(E, T, EF), E0, E1, Estep)
        return N

    def Nh(self, T, EF, E0, E1, Estep):
#        print("Nh: E range={:8.4g} - {:8.4g}  DOS={:12.4g}-{:12.4g}".format(E0, E1, self.DOSfh(E0, T, EF), self.DOSfh(E1, T, EF)))
#        print("     DOSfh:", self.DOS(E0), fh(E0, T, EF), fe(E0, T, EF))
#        print("     dE/kT=", (self.EV - E0) * e / kB / T, (E1 - self.EC) * e / kB / T)
        N = integrate_Simpson(lambda E: self.DOSfh(E, T, EF), E0, E1, Estep)
        return N

# Calculate NV and NC from DOS
    def NC_from_DOS(self, T, E0, E1, Estep):
        k = e / kB / T
        NC = integrate_Simpson(lambda E: self.DOS(E) * exp(-k * (E - self.EC)), E0, E1, Estep)
        return NC

    def NV_from_DOS(self, T, E0, E1, Estep):
        k = e / kB / T
        NV = integrate_Simpson(lambda E: self.DOS(E) * exp(-k * (self.EV - E)), E0, E1, Estep)
        return NV

# ionized donor density
    def NDp(self, EF, T):
        return self.ND * (1.0 - fe(self.ED, T, EF))

# ionized acceptor density
    def NAm(self, EF, T):
        return self.NA * fe(self.EA, T, EF)

# Calculate EF base on 
# the charge neutrality (electron number) condition, Ne(T, EF, dE) - N = 0,
# using the Newton method with the initial value EF0
    def FindEF_metal(self, T, EF0, E0, E1, Estep, totNe, hEF = 1.0e-7, dump = 1.0, maxiter = 50, ftol = 1.0e-3, EFtol = 1.0e-3, print_level = 0):
#        EF = optimize.newton(lambda EF: log(self.Ne(T, EF, E0, E1, Estep) + 1.0e-30) - log(totNe + 1.0e-30), EF0, x1 = EF0 + 1.0e-3, rtol = 1.0e-5, maxiter = 50)
        EF = EF0
        for i in range(maxiter):
            f  = log(self.Ne(T, EF,       E0, E1, Estep) + 1.0e-30) - log(totNe + 1.0e-30)
            if abs(f) < ftol:
                break

            fh = log(self.Ne(T, EF + hEF, E0, E1, Estep) + 1.0e-30) - log(totNe + 1.0e-30)
            f1 = (fh - f) / hEF
            delta_EF = - f / (f1 + dump)
            if abs(delta_EF) < EFtol:
                break
                
            EFnew = EF + delta_EF
            if print_level:
                print(f"iter {i}: f={f}   EF={EF} + {delta_EF} => {EFnew}")
            EF = EFnew
        else:
            print("")
            print("Error in tkDOS_VASP.FindEF_metal(): Did not converge with criteria dEF={:10.4g}<{} and S={:10.4g}<{:10.4g}".format(delta_EF, EFtol, f, ftol))
            print("")
            exit()

        return EF

    def FindEF_semi(self, T, EFmin, EFmax, E0h, E1h, E0e, E1e, Estep, dQ0 = 0.0,print_level = 1):
#        print_level = 2
# まず、EFmin,EFmaxにおけるΔQを計算し、それぞれが正・負あるいは負・生となっていることを確認する    
#        print("E range:", E0h, E1e)
        Nhmin, Nemin, NAmmin, NDpmin, RHmin, Nsmin = self.cal_carrier_densities(T, EFmin, E0h, E1h, E0e, E1e, Estep)
        dQmin  = Nemin + NAmmin - Nhmin - NDpmin - dQ0
        Nhmax, Nemax, NAmmax, NDpmax, RHmax, Nsmax = self.cal_carrier_densities(T, EFmax, E0h, E1h, E0e, E1e, Estep)
        dQmax = Nemax + NAmmax - Nhmax - NDpmax - dQ0
        iT = 0
        if print_level >= 1:
            print("        min (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                    .format(iT, EFmin, Nemin, Nhmin, NDpmin, NAmmin, dQmin))
            print("        max (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                    .format(EFmax, Nemax, Nhmax, NDpmax, NAmmax, dQmax))
#        dQmin = Ne(EFmin, T) + NAm(EFmin, T) - Nh(EFmin, T) - NDp(EFmin, T) - dQ0
#        dQmax = Ne(EFmax, T) + NAm(EFmax, T) - Nh(EFmax, T) - NDp(EFmax, T) - dQ0
#        print("  EFmin = {:12.8f}  dQmin = {:12.4g}".format(EFmin, dQmin))
#        print("  EFmax = {:12.8f}  dQmax = {:12.4g}".format(EFmax, dQmax))
        if print_level >= 1:
            print("    dQ=", dQmin, dQmax)
            print("    dQmin*dQmax=", dQmin * dQmax)
        if dQmin * dQmax > 0.0:
            print("Error: Initial Emin and Emax should be chosen as dQmin * dQmax < 0")
            return None, None

# 2分法開始
#        print_level = 2
        for i in range(self.nmaxiter_bisec):
            EFhalf = (EFmin + EFmax) / 2.0
            Nhh, Neh, NAmh, NDph, RHh, Nsh = self.cal_carrier_densities(T, EFhalf, E0h, E1h, E0e, E1e, Estep)
            if Nhh is None:
                return None, None

            dQhalf = Neh + NAmh - Nhh - NDph - dQ0
            if print_level >= 2:
                print("  iter {:03d}: half (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                        .format(i, EFhalf, Neh, Nhh, NDph, NAmh, dQhalf))
# EFの精度がepsより小さくなったら計算終了
            if abs(dQhalf) < 1.0:
                if print_level >= 2:
                    print("  Success: Convergence reached at EF = {} with small dQ={}".format(EFhalf, dQhalf))
                break
            if abs(EFmin - EFhalf) < self.eps_bisec and abs(EFmax - EFhalf) < self.eps_bisec:
                if print_level >= 2:
                    print("  Success: Convergence reached at EF = {}".format(EFhalf))
                break
            if dQmin * dQhalf < 0.0:
                EFmax = EFhalf
                dQmax = dQhalf
            else:
                EFmin = EFhalf
                dQmin = dQhalf
        else:
            print("  Failed: Convergence did not reach")
            return None, None

        return EFhalf, dQhalf

    def cal_carrier_densities(self, T, EF, E0h, E1h, E0e, E1e, Estep):
#        print("T={:8.4g} EF={:8.4g}   EV={:8.4g} E range h: {:8.4g} - {:8.4g}".format(T, EF, self.EV, E0h, E1h))
#        print("                       EC={:8.4g} E range h: {:8.4g} - {:8.4g}".format(self.EC, E0e, E1e))
        Nh  = self.Nh(T, EF, E0h, E1h, Estep)
        Ne  = self.Ne(T, EF, E0e, E1e, Estep)
        NAm = self.NAm(EF, T)
        NDp = self.NDp(EF, T)
#        print("N=", Nh, Ne)
        if Nh + Ne < 1.0:
            RH = 1.0e50
            Ns = 1.0
        else:
            RH = 1.0 / e * (Nh - Ne) / pow(Nh + Ne, 2.0)
            Ns = 1.0 / e / RH
        
        return Nh, Ne, NAm, NDp, RH, Ns

    def cal_EF_lists(self, T, xEF, print_level = 0):
        nEF    = len(xEF)
        yNe    = []
        yNh    = []
        yNDp   = []
        yNAm   = []
        yRH    = []
        yNs    = []
        yNsabs = []

        EFprev = self.EF0
        kBTe = kB * T / e
        dEinteg = self.nrange * kBTe
        E0h = self.EV - dEinteg
        E1e = self.EC + dEinteg
#    print("dEinteg=", dEinteg, T0)
#    exit()

        if print_level >= 1:
            print("")
            print("at {} K".format(T0))
            print ("%5s\t%8s\t%14s\t%14s\t%14s\t%14s\t%14s" % ("EF(eV)", "Ne", "Nh", "ND+", "NA-", "RH", "Ns"))
        for iEF in range(nEF):
            EF = xEF[iEF]
            Nh1, Ne1, NAm1, NDp1, RH1, Ns1 = self.cal_carrier_densities(T, EF, E0h, self.EV, self.EC, E1e, self.Estep)
            yNe.append(Ne1)
            yNh.append(Nh1)
            yNDp.append(NDp1)
            yNAm.append(NAm1)
            yRH.append(RH1)
            yNs.append(Ns1)
            yNsabs.append(abs(Ns1))
            if print_level >= 1:
                print ("%10.4f\t%12.6g\t%12.6g\t%12.6g\t%12.6g\t%12.6g\t%12.6g" % (EF, Ne1, Nh1, NDp1, NAm1, RH1, Ns1))

        return yNh, yNe, yNAm, yNDp, yRH, yNs, yNsabs

    def cal_T_lists(self, xT, dQ0 = 0.0, print_level = 0):
        nT    = len(xT)
        xTinv = []
        yEF   = []
        yNe    = []
        yNh    = []
        yNDp   = []
        yNAm   = []
        yRH    = []
        yNs    = []
        yNsabs = []
        EFprev = self.EF0
#初期範囲
        EFmin = min([self.EV, self.EF0 - self.dE_bisec])
        EFmax = max([self.EC, self.EF0 + self.dE_bisec])
#    dEFbisec = 0.1

        if print_level >= 1:
            print("")
            print ("%5s\t%8s\t%14s" % ("T(K)", "EF(eV)", "Ne(0 K) (check)"))
        for iT in range(nT):
            T = xT[iT]

            if print_level >= 1:
                print("")
                print("iT {:03d}: {} K".format(iT, T))

            kBTe = kB * T / e
            dEinteg = self.nrange * kBTe

            E0h = self.EV - dEinteg
            E1e = self.EC + dEinteg

            if E0h < self.Emin or self.Emax < E1e:
                print("*** Integration range error: [{}, {}] exceeds DOS E range [{}, {}]".format(E0h, E1e, Emin, Emax))
                exit()

#        Ne0K = self.Ne(0.0, self.EF0, E0h, self.EF0, self.Estep)
            EF, dQ = self.FindEF_semi(T, EFmin, EFmax, E0h, self.EV, self.EC, E1e, self.Estep, dQ0 = dQ0)
            Nhh, Neh, NAmh, NDph, RH, Ns = self.cal_carrier_densities(T, EF, E0h, self.EV, self.EC, E1e, self.Estep)

            xTinv.append(1000.0 / T)
            yEF.append(EF)
            yNe.append(Neh)
            yNh.append(Nhh)
            yNDp.append(NDph)
            yNAm.append(NAmh)
            yRH.append(RH)
            yNs.append(Ns)
            yNsabs.append(abs(Ns))
            if print_level >= 1:
                print("  ***(T,EF,Ne,Nh,NDp,NAm,RH)=({:6.4f},{:6.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                            .format(T, EF, Neh, Nhh, NDph, NAmh, RH, Ns))

        EFprev = EF
        EFmin = EF - self.dE_bisec
        EFmax = EF + self.dE_bisec
        print ("%5.0f\t%12.6g" % (T, EF))

        return xTinv, yEF, yNe, yNh, yNAm, yNDp, yRH, yNs, yNsabs

def main():
    print("")
    print("This is a library, not runnable")
    print("")


if __name__ == "__main__":
    main()
