from math import exp, sqrt, log
import numpy as np
from pprint import pprint
import openpyxl

from tklib.tkobject import tkObject
from tklib.tkfile import tkFile
from tklib.tkinifile import tkIniFile
from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, analyze_varstr
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me
#from tklib.tksci.tkoptimize import mlsq_general


ekB = e / kB


def split_optstr(str):
    try:
        val, id = str.split(':')
        return pfloat(val), pint(id)
    except:
        return str, None

def strlist(vars, format):
    s = ""
    for i in range(len(vars)):
       s += format.format(vars[i]) + ', '
    return s


class tkMobility(tkObject):
    def __init__(self):
        super(tkObject, self).__init__()

        self.path   = None
        self.header = None
        self.xT     = None
        self.yN     = None
        self.ymu    = None
        self.ys     = None

# 電荷、有効質量
        self.charge = None          # in e, 1.0 for hole, -1.0 for electron
        self.meff   = None          # in me
# 散乱時間パラメータ。散乱因子 r、平均自由行程のprefactor
        self.rfac   = None          # tau = (meff/2)^0.5 * l0(T) * E^(r-0.5)
        self.l0     = None          # m
        self.l0norm = None          # m

        self.varkeys = None
        self.varname = None
        self.varunit = None
        self.ai0     = None         # Default values
        self.ai      = None
        self.optid   = None

        self.debug   = 0
        self.use_simple = 0
        
    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()

    def set(self, **args):
        self.update(args)

    def set_scattering_parameters(self, l0 = None, rfac = None):
        if l0 is None:
            l0 = self.l0
        if rfac is None:
            rfac = self.rfac

        self.l0 = l0
        self.rfac = rfac
        self.l0norm = l0 / pow(e, rfac)

    def tau(self, T, E):
        if E <= 0.0:
            return None
        return sqrt(self.meff * me * 0.5) * self.l0norm * pow(E * e, self.rfac - 0.5) # s, E in eV
    
    def recover_parameters(self, ais, optid = None, aidef = None):
        if optid is None:
            optid = self.optid
        if aidef is None:
            aidef = self.ai0

        ai = []
        c = 0
        for i in range(len(optid)):
            if optid[i] == 1:
                ai.append(ais[c])
                c += 1
            else:
                ai.append(aidef[i])

        return ai

    def choose_parameters(self, ais0, optid = None):
        if optid is None:
            optid = self.optid

        ais = ais0.copy()
#ai0     = [mu0, Eb, a1, a2]
    
        ai = []
        c = 0
        for i in range(len(optid)):
            if optid[i] == 1:
                ai.append(ais[i])

        return ai

    def parameter_list(self):
        list = []
        nvars = len(self.varkeys)
        for i in range(nvars):
            key = self.varkeys[i]
            list.append(self.get2(key))
        return list
    
    def parameter_tuple(self):
        return tuple(self.parameter_list())

    def set_parameters(self, ai = None):
        if ai is None:
            ai = self.ai

        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
#            print("set: key=", key, ai[i])
            self.set_attribute2(key, ai[i])

    def read_parameters(self, path, optid = None):
        if optid is None:
            optid = self.optid

        ini = tkIniFile()
        inf = ini.ReadAll(path, AddSection = 0)
        if inf is None:
            return

        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            str = inf.get(key, None)
            val, id = split_optstr(str)
#            print("read: ", key, val, id)
            if val is not None:
                self.set_attribute2(key, pfloat(val))
            if id is not None:
                optid[i] = id
        
    def save_parameters(self, path, optid = None):
        if optid is None:
            optid = self.optid

        ini = tkIniFile(path)
        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            val = self.get2(key, None)
            if val is not None:
#                print("save key=", key, val, optid[i])
                ini.WriteString('Parameters', key, "{}:{}".format(val, optid[i]))

    def print_parameters(self, varname = None, varunit = None, ai = None, optid = None):
        if varname is None:
            varname = self.varname
        if varunit is None:
            varunit = self.varunit
        if ai is None:
            ai = self.parameter_list()
        if optid is None:
            optid = self.optid

        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            val = self.get2(key, None)
            if val is not None:
                print("  {:10s}: {:14.8g}  optid={}".format(key, pfloat(val), optid[i]))
            else:
                print("  {:10s}: None  optid={}".format(key, optid[i]))

    def read_excel(self, infile):
        wb = openpyxl.load_workbook(infile, data_only = True)
        if not wb:
            print("")
            print("Error to read [{}]".format(infile))
            print("")
            terminate()

        sheetnames = wb.sheetnames
        print("sheet names:", sheetnames)
        print("  read [{}]".format(sheetnames[0]))

#       ws = wb.active
        ws = wb[sheetnames[0]]
        Tlabel  = ws[1][0].value
        Nlabel  = ws[1][1].value
        mulabel = ws[1][2].value
        slabel  = ws[1][3].value
        print("Labels:", Tlabel, Nlabel, slabel, mulabel)
        xT  = []
        yN  = []
        ymu = []
        ys  = []
        i = 2
        while 1:
            Tcell = ws.cell(row = i, column = 1).value
            if Tcell is None or Tcell == '':
                break

            Ncell  = ws.cell(row = i, column = 2).value
            mucell = ws.cell(row = i, column = 3).value
            scell  = ws.cell(row = i, column = 4).value
#            print("cells=", Tcell, Ncell, scell, mucell)

            Tcell  = float(Tcell)
            Ncell  = float(Ncell)
            scell  = float(scell)
            mucell = float(mucell)

            xT.append(Tcell)
            yN.append(Ncell)
            ys.append(scell)
            ymu.append(mucell)
            i += 1

        self.header = [Tlabel, Nlabel, mulabel, slabel]
        self.xT  = xT
        self.yN  = yN
        self.ymu = ymu
        self.ys  = ys

        return self.header, xT, yN, ymu, ys


def main():
    print("")
    print("This is a library, not runnable")
    print("")


if __name__ == "__main__":
    main()
