import numpy as np
from numpy import exp
import pandas as pd
from scipy.optimize import minimize


method = "nelder-mead"


#==========================================
# Source parameters to be fitted
#==========================================
infile = 'peak.xlsx'
nx = None

# optimization parameters
func = '2.0 * (x[0] - 3.0)**2 + (x[1] - 1.0)**4 + 2.0'
x0s = [1.3,  0.6, 0.1]

maxiter = 100
tol = 1.0e-5


#==========================================
# functions
#==========================================
def cal_y(x, params):
    I0 = params[0]
    x0 = params[1]
    w  = params[2]
    a = 0.832554611 / w
    X = a * (x - x0)
    return I0 * exp(-X * X)

def ycal_list(xd, params):
    y = []
    for i in range(len(xd)):
        y.append(cal_y(xd[i], params))
    return y

def minimize_func(xk, xd, yd):
    nd = len(xd)
    MSE = 0.0
    for i in range(nd):
        yc = cal_y(xd[i], xk)
        d  = yd[i] - yc
        MSE += d * d

    return MSE / nd


#==========================================
# Main routine
#==========================================
def main():
    print("")
    print("Simplest peak fitting program by simplex")
    print(f"method        : {method}")
    print(f"input file    : {infile}")
    print(f"initial values: {x0s}")
    print(f"tol           : {tol}")
    print(f"maxiter       : {maxiter}")

    print("")
    print(f"Read [{infile}]")
    df = pd.read_excel(infile, engine = 'openpyxl')
    labels = df.columns.to_list()
    xd = df[labels[0]].to_numpy()
    yd = df[labels[1]].to_numpy()
    nx = len(xd)
    xrange = [min(xd), max(xd)]

    res = minimize(lambda xk: minimize_func(xk, xd, yd), x0s, method = method, tol = tol, options = {'maxiter':maxiter, "disp":True})
#    print("")
#    print(res)
    print("")
    if res.success:
        print(f"Function MSE takes the minimum")
        print(f"   at MSE={res.fun}")
        print(f"   with x={res.x}")
        print(f" iteration: {res.nit}")
    else:
        print(f"Function did not converge")
        print(res)


if __name__ == "__main__":
    main()
